
#include <functional>
#include <iostream>

struct MyType {
  int val = 0;
  void incval(int i) {
    val = val + i;
    std::cout << "val is " << val << std::endl;
  }
  std::function<void()> make_incrementer(int increment) {
    // Note this is immutable (MyType *const), but *this can be updated
    return [increment,this]() { incval(increment); };
  }
};

int
main()
{
  std::function<int(int,int)> adder = [](int a, int b) {
    return a + b;
  };
  std::cout << "adder(1,2) -> " << adder(1,2) << std::endl;

  int x = 7, y = 1;
  auto fn = [&x, y]() { x += y; };
  y = 1000;
  fn();

  [x]() {
    std::cout << "x is " << x << std::endl;
    //x++;   // <- an error, because x is read-only copy of x in main
  }();
  std::cout << "x is still " << x << std::endl;

  [&x]() {
    // now we have captured x by reference
    std::cout << "x is " << x++ << std::endl;
  }();
  std::cout << "x is now " << x << std::endl;

  MyType mt;
  std::function<void()> inc2 = mt.make_incrementer(2);
  inc2();
  inc2();

  [&]() {	  // plain & captures everything in scope by reference
    x++;
    inc2();
  }();

  [=,&x]() { // = captures everything by copy, except x by reference
    x++;
    //mt.incval(1); // error, mt is const, but incval() not const
    inc2();	    // okay, inc2 is const, but has pointer to mutable mt
  }();
}
