
#include <chrono>
#include <iostream>
#include <string>

#include <poll.h>

#include "stack.hh"
#include "timer.hh"

using namespace std::string_literals;

using Bytes = std::unique_ptr<char[]>;

struct Threadlet {
  std::string name;
  sp_t sp;
  Bytes stack;
  Threadlet *next;

  void run();
};

Threadlet *curthread = new Threadlet{"initial thread"};

void
Threadlet::run()
{
  Threadlet *prev = curthread;
  curthread = this;
  stack_switch(&prev->sp, &this->sp);
}

std::atomic_int good_counter;
volatile int bad_counter;

void
delay(int msec)
{
  using clock = std::chrono::steady_clock;
  clock::time_point start = clock::now();
  while (clock::now() - start < std::chrono::milliseconds(msec)) {
    int old_value = bad_counter;
    ++good_counter;
    bad_counter = old_value+1;
  }
}

void
worker()
{
  intr_enable(true);		// <-- Why?
  for (int i = 0; i < 30; ++i) {
    delay(30);
    std::cout << "running in " << curthread->name << std::endl;
  }

  std::cout << "good counter = " << good_counter
	    << ", bad counter = " << bad_counter << std::endl;
  exit(0);
}

void
timer_handler()
{
  // Note: interrupts are disabled here
  curthread->next->run();
  // Interrupts will be restored when this function returns
}

int
main()
{
  Threadlet *t = curthread;
  for (int i = 0; i < 3; ++i) {
    t = new Threadlet{
      "thread #"s + std::to_string(i), nullptr, Bytes(new char[8192]), t
    };
    t->sp = stack_init(t->stack.get(), 8192, worker);
  }
  curthread->next = t;

  timer_init(100'000, timer_handler);
  worker();
}
