
#include <atomic>
#include <sstream>
#include <thread>

#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>

void
delay(int msec)
{
  using clock = std::chrono::steady_clock;
  clock::time_point start = clock::now();
  while (clock::now() - start < std::chrono::milliseconds(msec))
    ;
}


std::atomic_int counter;

void
count(const char *name)
{
  for (int i = 0; i < 10; ++i) {
    delay(10);

    std::ostringstream out;
    out << name << ": counter is " << counter++ << std::endl;
    std::string s = out.str();

    // Writes of less than PIPE_BUF bytes are atomic.  Gatherine
    // everything into a string and issuing a single write system call
    // avoids intermigling output from multiple threads/processes on
    // the same line.
    write(1, s.c_str(), s.size());
  }
  close(1);
}

int
main()
{
#if 1
  pid_t p1, p2;
  if (!(p1 = fork())) {
    count("process p1");
    exit(0);
  }
  if (!(p2 = fork())) {
    count("process p2");
    exit(0);
  }
  waitpid(p1, nullptr, 0);
  waitpid(p2, nullptr, 0);

#else
  std::thread t1([]() { count("thread t1"); });
  std::thread t2([]() { count("thread t2"); });
  t1.join();
  t2.join();
#endif
}
