
#include <iostream>
#include <string>

#include "stack.hh"

using namespace std::string_literals;

using Bytes = std::unique_ptr<char[]>;

struct Threadlet {
  std::string name;
  sp_t sp;
  Bytes stack;
  Threadlet *next;

  void run();
};

// Note how we leave stack NULL for the initial thread (already have
// one stack when program starts; can't free that initial stack).
Threadlet *curthread = new Threadlet{"initial thread"};

void
Threadlet::run()
{
  Threadlet *prev = curthread;
  curthread = this;
  stack_switch(&prev->sp, &this->sp);
}

void
worker()
{
  for (;;) {
    std::cout << "running in " << curthread->name << std::endl;
    curthread->next->run();
  }
}

int
main()
{
  Threadlet *t = curthread;
  for (int i = 0; i < 3; ++i) {
    t = new Threadlet{
      "thread #"s + std::to_string(i), nullptr, Bytes(new char[8192]), t
    };
    t->sp = stack_init(t->stack.get(), 8192, worker);
  }
  curthread->next = t;

  curthread->next->run();
  std::cout << "back in " << curthread->name << std::endl;
  curthread->next->run();
}
